#define LIB_NAME "thread"
#define MODULE_NAME "thread"

#include <dmsdk/dlib/thread.h>
#include <dmsdk/sdk.h>
//lualib.h no exposed by defold, what can one do, but get desperate
#include "math.c"
#include "string.c"
#include "table.c"

struct Context{

    bool finish;
    bool running;
    const char* result;
    const char* callback; 
    
    const char* arg;
    const char* worker;
    const char* chunk;
};

Context global_data;


static void stackDump(lua_State* lua) {
    if(lua == NULL){
        dmLogInfo("-----lua_State* is null-------- ");
        return;
    }
    int i, t;
    int top = lua_gettop(lua);
    dmLogInfo("-----STACK DUMP -------- ")
    for (i = 1; i <= top; i++) {
        t = lua_type(lua, i);
        switch (t) {
            case LUA_TSTRING:
                dmLogInfo("%d : \"%s\"", i, lua_tostring(lua, i));
                break;
            case LUA_TBOOLEAN:
                dmLogInfo("%d : %s", i, lua_toboolean(lua, i) ? "true" : "false");
                break;
            case LUA_TNUMBER:
                dmLogInfo("%d : %d ", i, lua_tonumber(lua, i));
                break;
            default:
                dmLogInfo("%d : Type(%s)", i, lua_typename(lua, i));
                break;
        }
    }
    dmLogInfo("----------------------");
}

static int myPrint(lua_State* lua){
    const char* str = lua_tostring(lua, -1);
    dmLogInfo("%s", str);
    return 0;
}

static const luaL_reg mylibs_methods[] =
{
    {"log", myPrint},
    {0, 0}
};

static void Worker(void* _ctx)
{
   
    //dmLogInfo("-------------Running Worker---------");
    lua_State* t = luaL_newstate();
    //luaL_openlibs(t);

    luaL_register(t, "mylibs", mylibs_methods);
    lua_pop(t, 1);

    luaL_register(t, "math", mathlib);
    lua_pop(t, 1);

    luaL_register(t, "string", strlib);
    lua_pop(t, 1);

    luaL_register(t, "table", tab_funcs);
    lua_pop(t, 1);
    
    
    int status = luaL_dostring(t , global_data.chunk);
    lua_getfield(t, LUA_GLOBALSINDEX, global_data.worker);
    lua_pushstring(t, global_data.arg);
    
    //dmLogInfo("------------New Thread Stack Dump ---------------");
    //stackDump(t);

    
    int error = lua_pcall(t, 1, 1, 0);
    dmLogInfo("Executed %s, Result code:%d" , global_data.worker,  error); 
    //dmLogInfo("------------New Thread Stack Dump After Resume ---------------");
    //stackDump(t);
    
    if (lua_type(t, -1) == LUA_TSTRING) {
        global_data.result = lua_tostring(t, -1); //return result
    }
    
    global_data.finish = true;
    //dmLogInfo("----Thread Done----");
}

static int RunThread(lua_State* L){

    if (global_data.running){
        dmLogInfo("Error:Another thread is running");
        return 0;
    }
    
    int top = lua_gettop(L);

    //dmLogInfo("About to execute worker thread");
    //stackDump(L);
    
    for (int i = 1 ; i < 5 ; i++){
        if (lua_type(L, -1) != LUA_TSTRING) {
            stackDump(L);
            dmLogError("Expected: String %g", i);
            return 0;
        }
    }

    global_data.chunk    = lua_tostring(L, 1);
    global_data.arg      = lua_tostring(L, 2);
    global_data.worker   = lua_tostring(L, 3);

    global_data.result   = "";
    global_data.callback = lua_tostring(L, 4);
    global_data.finish   = false;

    
    dmThread::Thread thread = dmThread::New(Worker, 0x80000, 0, "defold_thread");
    dmLogInfo("Thread running.");
    global_data.running = true;
    
    return 0;
}


// Functions exposed to Lua
static const luaL_reg Module_methods[] =
{
    {"run", RunThread},
    {0, 0}
};

static void LuaInit(lua_State* L)
{
    int top = lua_gettop(L);

    // Register lua names
    luaL_register(L, MODULE_NAME, Module_methods);
    lua_pop(L, 1);
    
    assert(top == lua_gettop(L));
}

dmExtension::Result UpdateThread(dmExtension::Params* params)
{
    if (global_data.running && global_data.finish) {
        dmLogInfo("Thread finish: calling %s", global_data.callback);
        lua_getfield(params->m_L,LUA_GLOBALSINDEX, global_data.callback);
        lua_pushstring(params->m_L, global_data.result);
        lua_pcall(params->m_L, 1, 0, 0);
        
        global_data.running = false;
    }
    return dmExtension::RESULT_OK;
}

dmExtension::Result AppInitializeThread(dmExtension::AppParams* params)
{
    return dmExtension::RESULT_OK;
}

dmExtension::Result InitializeThread(dmExtension::Params* params)
{
    global_data.running = false;
    global_data.finish  = false;
    
    // Init Lua
    LuaInit(params->m_L);
    printf("Registered %s Extension\n", MODULE_NAME);
    return dmExtension::RESULT_OK;
}

dmExtension::Result AppFinalizeThread(dmExtension::AppParams* params)
{
    return dmExtension::RESULT_OK;
}

dmExtension::Result FinalizeThread(dmExtension::Params* params)
{
    return dmExtension::RESULT_OK;
}


// Defold SDK uses a macro for setting up extension entry points:
//
// DM_DECLARE_EXTENSION(symbol, name, app_init, app_final, init, update, on_event, final)

// MyExtension is the C++ symbol that holds all relevant extension data.
// It must match the name field in the `ext.manifest`
DM_DECLARE_EXTENSION(thread, LIB_NAME, AppInitializeThread, AppFinalizeThread, InitializeThread, UpdateThread, 0, FinalizeThread)

